/*
 *  imp_lwo.c - LightWave import plugin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *  (C) 2002 Markus Dahms <dahms@fh-brandenburg.de>
 *
 *  based on gtkglarea example viewlw:
 *  Copyright (C) 1998 Janne Lf <jlof@mail.student.oulu.fi>
 *
 */

#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include <g3dviewer_plugin.h>

#define LW_MAX_POINTS   200
#define LW_MAX_NAME_LEN 500

int lwo_read_directory(FILE *f, int nbytes, G3DModel *model, void *pobject,
                        guint32 parentid, int level);

/*****************************************************************************/
/* plugin interface                                                          */
/*****************************************************************************/

int plugin_load(const char *filename, G3DModel *model)
{
  G3DObject *lwo_object;
  FILE *f;
  guint32 id, len;

  g_return_val_if_fail(model != NULL, EXIT_FAILURE);

  model->lights = NULL;
  model->objects = NULL;

  f = g3d_iff_open(filename, &id, &len);
  if(f == NULL)
  {
    g_warning("can't open file '%s'", filename);
    return EXIT_FAILURE;
  }

  if(id != G3D_IFF_MKID('L','W','O','B'))
  {
    g_warning("file '%s' is not a LightWave object", filename);
    fclose(f);
    return EXIT_FAILURE;
  }

  g3d_interface_progress_init(model);

  lwo_object = g_new0(G3DObject, 1);
  lwo_object->name = g_strdup("LightWave object");
  model->objects = g_slist_append(model->objects, lwo_object);

  lwo_read_directory(f, len, model, lwo_object, id, 1);

  fclose(f);

  g3d_interface_progress_finish(model);

#if DEBUG > 2
  g_printerr("DEBUG: lwo file loaded\n");
#endif

  return EXIT_SUCCESS;
}

char *plugin_description(void)
{
  return g_strdup(
    "Import plugin to load LightWave Objects (.lwo files)\n"
    "Author: Markus Dahms, heavily copied from gtkglarea examples\n");
}

char **plugin_extensions(void)
{
  return g_strsplit("lwo:lwb:lw", ":", 0);
}

/*****************************************************************************/
/* private                                                                   */
/*****************************************************************************/

gint lwo_read_string(FILE *f, char *s);
void lwo_read_srfs(FILE *f, gint nbytes, G3DObject *object);
void lwo_read_surf(FILE *f, gint nbytes, G3DObject *object);
void lwo_read_pols(FILE *f, gint nbytes, G3DObject *object);
void lwo_read_pnts(FILE *f, gint nbytes, G3DObject *object);

int lwo_read_directory(FILE *f, int nbytes, G3DModel *model, void *pobject,
                       guint32 parentid, int level)
{
#if DEBUG > 0
  int i;
#endif
  int toread = nbytes;
  float pscale = (float)toread / 100.0;
  while(toread > 0)
  {
    guint32 id, len;

    if(level == 0) 
    {
      g3d_interface_progress_update(model, 100.0 - (float)toread / pscale);
    }

    toread -= g3d_iff_readchunk(f, &id, &len);
#if DEBUG > 0
    for(i=0; i<level; i++) g_printerr("  ");
    g_printerr("[%c%c%c%c] (%d bytes)\n", 
               (id >> 24) & 0xFF, (id >> 16) & 0xFF,
               (id >> 8) & 0xFF, id & 0xFF, len);
#endif
    switch(id)
    {
      case G3D_IFF_MKID('P','N','T','S'):
        lwo_read_pnts(f, len, (G3DObject*)pobject);
        break;
      case G3D_IFF_MKID('P','O','L','S'):
        lwo_read_pols(f, len, (G3DObject*)pobject);
        break;
      case G3D_IFF_MKID('S','R','F','S'):
        lwo_read_srfs(f, len, (G3DObject*)pobject);
        break;
      case G3D_IFF_MKID('S','U','R','F'):
        lwo_read_surf(f, len, (G3DObject*)pobject);
        break;

      default:
        fseek(f, len + (len%2), SEEK_CUR);
    }
  }
  return EXIT_SUCCESS;
}

/*****************************************************************************/
/* LWO specific                                                              */
/*****************************************************************************/

gint lwo_read_string(FILE *f, char *s)
{
  gint c;
  gint cnt = 0;
  do {
    c = g3d_read_int8(f);
    if (cnt < LW_MAX_NAME_LEN)
      s[cnt] = c;
    else
      s[LW_MAX_NAME_LEN-1] = 0;
    cnt++;
  } while (c != 0);
  /* if length of string (including \0) is odd skip another byte */
  if (cnt%2) {
    g3d_read_int8(f);
    cnt++;
  }
  return cnt;
}

void lwo_read_srfs(FILE *f, gint nbytes, G3DObject *object)
{
#if DEBUG > 2
  g_printerr("lwo_read_srfs {");
#endif
  while(nbytes > 0)
  {
    char text[1024];
    G3DMaterial *material = g3d_new_G3DMaterial();
    
    nbytes -= lwo_read_string(f, text);
    material->name = g_strdup(text);
    object->materials = g_slist_append(object->materials, material);
  }
#if DEBUG > 2
  g_printerr(" }\n");
#endif
#if 0
  g_printerr("materials: %p\n", object->materials);
#endif
}

void lwo_read_surf(FILE *f, gint nbytes, G3DObject *object)
{
  GSList *mlist;
  G3DMaterial *material = NULL, *tmat;
  char name[LW_MAX_NAME_LEN];

#if DEBUG > 2 
  g_printerr("lwo_read_surf {");
#endif
  nbytes -= lwo_read_string(f, name);

  mlist = object->materials;
  g_return_if_fail(mlist != NULL);
  while(mlist != NULL)
  {
    tmat = (G3DMaterial*)mlist->data;
#if 0
    g_printerr("name: %s\n", tmat->name);
#endif
    if(strcmp(name, tmat->name) == 0)
    {
      material = tmat;
      break;
    }
    mlist = mlist->next;
  }
  g_return_if_fail(material != NULL);
#if 0
  g_printerr("name: %s\n", material->name);
#endif
  while(nbytes > 0)
  {
    gint32 id = g3d_read_int32_be(f);
    gint32 len = g3d_read_int16_be(f);
    
    nbytes -= 6 + len + (len%2);

    switch(id)
    {
      case G3D_IFF_MKID('C','O','L','R'):
        material->r = g3d_read_int8(f) / 255.0;
        material->g = g3d_read_int8(f) / 255.0;
        material->b = g3d_read_int8(f) / 255.0;
        g3d_read_int8(f);
        break;
      default:
        fseek(f, len+(len%2), SEEK_CUR);
    }
  }
#if DEBUG > 2
  g_printerr(" }\n");
#endif
}

void lwo_read_pols(FILE *f, gint nbytes, G3DObject *object)
{
  int nb = nbytes, n = 0;
#if DEBUG > 2
  g_printerr("lwo_read_pols {");
#endif
  while(nb > 0)
  {
    G3DFace *face = g_new0(G3DFace, 1);
    int i, nmat;

    n++;
    face->num_vertices = g3d_read_int16_be(f);
    face->index_vertex = g_malloc0(face->num_vertices * sizeof(int));
    nb -= 2;
    
    for(i=0; i<face->num_vertices; i++)
    {
      face->index_vertex[i] = g3d_read_int16_be(f);
      nb -= 2;
      g3d_interface_update();
    }

    nmat = g3d_read_int16_be(f);
    nb -= 2;

    if(nmat < 0) 
    {
      /* detail polygons, skipped */
      int det_cnt = g3d_read_int16_be(f);
      nb -= 2;
      nmat *= -1;
      while(det_cnt-- > 0)
      {
        int cnt = g3d_read_int16_be(f);
        fseek(f, cnt*2+2, SEEK_CUR);
        nb -= cnt*2+2;
        g3d_interface_update();
      }
    }
    face->material = g_slist_nth_data(object->materials, (guint)nmat-1);
#if 0
    g_printerr("material: %p, mat: %d\n", face->material, nmat);
#endif
    object->faces = g_slist_append(object->faces, face);
#if DEBUG > 3
    g_printerr("lwo_read_pols: face #%d (nbytes: %d)\n", n, nb);
#endif
  }
#if DEBUG > 2
  g_printerr(" }\n");
#endif
}

void lwo_read_pnts(FILE *f, gint nbytes, G3DObject *object)
{
  int i;
#if DEBUG > 2
  g_printerr("lwo_read_pnts {");
#endif
  object->vertex_count = nbytes / 12;
  object->vertices = g_malloc0(sizeof(GLfloat) * object->vertex_count * 3);
  for(i=0; i<object->vertex_count; i++)
  {
    object->vertices[i*3+0] = g3d_read_float_be(f);
    object->vertices[i*3+1] = g3d_read_float_be(f);
    object->vertices[i*3+2] = g3d_read_float_be(f);
    g3d_interface_update();
  }
#if DEBUG > 2
  g_printerr(" }\n");
#endif
}


